"""
Plate
------
Primary functionality of ninetysix, providing the a class for the
rapid analysis and visualization of ninety-six* well plate assays.

*Not actually restricted to 96 wells, pretty much anything goes.
"""

import numpy as np
import pandas as pd

from .parsers import well_regex, pad, _infer_padding, _infer_casing
from .util import check_inputs, check_annotations, check_df_col
from .util._plate_attrs import _set_viz_attrs, _set_pandas_attrs


class Plate():
    """A pandas DataFrame-centric, well-value oriented container for
    structuring, annotating, and analyzing 96-well* plate data.

    (*Contrary to the package name, can be more or less than 96 wells.)

    Supports assignment of well conditions during and after object
    instantiation to result in the construction of condition-assigned,
    well-value pairs. Well column remains at 0 index, working value
    column remains -1 index to support rapid, implicit visualization of
    pertinent well-value data.

    Parameters:
    -----------
    data: dict, DataFrame, Iterable (list of length-two lists), or path
        Given a dict or DataFrame, the data must contain a key/column
        name that identifies 'well' (case-insentivie). The value of 
        interest can be assigned a name in the final DataFrame via
        `value_name` kwarg. Passing in an Iterable assumes tuples of
        `(well, value)` ordering and assigns as such. Passing a string
        will assume a path to a pandas-readable csv or excel file, using
        the default settings. If more control over the import of this
        file is needed, import to a pandas DataFrame first, then pass
        this to `data`.
    value_name: string or list
        Specification or assignment of value name(s). If given as a
        list, the last entry is used as the primary working value.
    annotate: nested dictionary or template excel file
        Maps wells to conditions in new columns of a tidy DataFrame. The
        outermost keys give the name of the resulting column. The inner
        keys are the wells corresponding to a given condition/value.
        Inner keys support simply regex specification of well, such as
        '[A-C,E]2' for 'A2', 'B2', 'C2', 'E2'. (Or 'A02', etc., which is
        inferred from the initial construction.) Inner keys can also be
        one of (default, standard, else, other), and the value of this
        key will be assigned to all other non-specified wells (else they
        get a value of `None`). Also takes a specific excel spreadsheet
        format; see the annotate_wells() method or  parsers.well_regex() function for more details.
    zero_padding: bool, default None
        Whether or not the wells are (or should be) zero-padded, e.g.,
        A1 or A01. If None, determines this from what's given. If True
        or False, will update the wells to match this state.
    case: str method, default str.lower
        Assigns the case of auto-generated DataFrame columns.
    pandas_attrs: bool, default True
        Whether or not to assign pandas attributes and methods directly
        to Plate object rather than only being accessible via the
        underlying Plate.df attribute. If it fails due to an import
        error, the attribute becomes a message (Failed due to import
        error).

    Examples:
    ---------
    Constructing a Plate object with 'data':

    # With an Iterable
    >>> input = zip(['A1', 'A2'], [1, 0.5])
    >>> ns.Plate(data=input, value_name='area')
        well    row    column     area
    0   'A1'    'A'         1        1
    1   'A2'    'A'         2      0.5

    # With a dictionary (or DataFrame)
    >>> input_dict = {
    ...     'well': ['A1', 'A2'],
    ...     'area': [1, 0.5],
    ...     'RT': [0.42, 0.41],
    ... }
    >>> ns.Plate(data=input_dict, value_name='area')
        well    row    column      RT    area
    0   'A1'    'A'         1    0.42       1
    1   'A2'    'A'         2    0.41     0.5

    # Force capitalized and with zero-padding
    >>> input_dict = {
    ...     'well': ['A01', 'A02'],
    ...     'Area': [1, 0.5],
    ...     'R.T.': [0.42, 0.41],
    ... }
    >>> ns.Plate(data=input_dict, value_name='Area', case=str.title)
        Well     Row   Column      R.T.     Area
    0   'A01'    'A'        1    0.42        1
    1   'A02'    'A'        2    0.41      0.5

    # Assign wells
    >>> input_dict = {
    ...     'well': ['A1', 'A2'],
    ...     'area': [1, 0.5],
    ...     'RT': [0.42, 0.41],
    ... }
    >>> controls = {
    ...     'controls': {
    ...         'A1': 'experiment',
    ...         'A2': 'negative',
    ...     }
    ... }
    >>> ns.Plate(data=input_dict, value_name='area', annotate=controls)
        well    row   column      RT       controls    area
    0   'A1'    'A'        1    0.42   'experiment'       1
    1   'A2'    'A'        2    0.41     'negative'     0.5
    """

    def __init__(
        self,
        data=None,
        value_name=None,
        case=None,
        zero_padding=None,
        annotate=None,
        # fill_plate=False,
        # n_wells=96,
        pandas_attrs=True,
    ):

###########################
# Initial set up
###########################
#         
        # Assign attributes
        self._data = data
        self._value_name = value_name
        self._zero_padding = zero_padding
        self._case = case

        # Convert data into initial DataFrame-like object
        self.data = data

        # Check that all inputs are acceptable
        self._passed = check_inputs(self)

        # Get information and regularize the dataframe
        self.df, self._well, self._row, self._col = self._generate_df()

        # Try to add pandas_attrs
        self._pandas_attrs = pandas_attrs
        if self._pandas_attrs:
            try:
                _set_pandas_attrs(self)
            except ImportError:
                self._pandas_attrs = 'Failed due to import error'

        # Add high-level attributes
        self._locations = [self._well, self._row, self._col]
        self._annotations = [
            col for col in self.df.columns
            if col not in [*self.locations, self.value_name]
        ]
        self._values = [self.value_name]

        # Pack into dictionary
        self._column_dict = {
            'locations': self.locations,
            'annotations': self.annotations,
            'values': self.values
        }

        # Organize DataFrame
        self.mi_df = self._generate_mi_df()
        self.zero_padding = zero_padding
        self.case = case

        # Annotate
        if annotate is not None:
            # Cannot reassign self in init, so need to be clever
            self.mi_df = self.annotate_wells(annotate).mi_df
            self.column_dict = self.annotate_wells(annotate).column_dict

        # Set plotting attributes directly as methods
        _set_viz_attrs(self)

###########################
# Operator Overloads
###########################

    def __getitem__(self, key):
        """Overwritten to get from DataFrame attribute"""
        return self.df[key]

    def __setitem__(self, key, value):
        """Overwritten to set to DataFrame object, as annotation"""
        key = self._set_case(key)
        
        # Add to df
        self.df[key] = value

        # Create series for alignment on mi_df and add
        series = self.df[[self._well, key]].set_index(self._well)
        self.mi_df[('annotations', key)] = series
        if key not in self.annotations:
            self.annotations = [*self.annotations, key]

    def __delitem__(self, key):
        """Overwritten to prohibit deleting base locations and to update
        the location, annotation, and value attributes.
        """
        base_locs = [self._well, self._row, self._col]
        if key in base_locs:
            raise ValueError(
                'Cannot delete base locations (well, row, column).'
            )
        del self.df[key]
        for attr in ('locations', 'annotations', 'values'):
            attr_list = getattr(self, attr)
            if key in attr_list:
                attr_list.remove(key)
                del self.mi_df[(attr, key)]

    def __len__(self):
        return len(self.df)

    def _repr_html_(self):
        """Sets HTML representation as DataFrame attribute"""
        return self.df._repr_html_()

################################
# Properties and support methods
################################

### Data
    @property
    def data(self):
        """Makes sure data is not zip, since it's called multiple times,
        and read in to initial pd.DataFrame.
        """
        return self._data

    @data.setter
    def data(self, data):
        # Possible well-value pair types
        wv_pair_types = (zip, list, tuple, np.ndarray)
        if isinstance(data, wv_pair_types):
            data = pd.DataFrame(
                list(data),
                columns=['well', None]
            )
        if isinstance(data, str):
            extension = data.split('.')[-1]
            if extension == 'csv':
                data = pd.read_csv(data)
            if extension in ('xls', 'xlsx'):
                engine = 'openpyxl' if extension == 'xlsx' else 'xlrd'
                data = pd.read_excel(data, engine=engine)
        if isinstance(data, dict):
            data = pd.DataFrame(data)
        if not isinstance(data, pd.DataFrame):
            raise NotImplementedError(
                'Data format not supported.'
            )
        self._data = data

### Casing
    @property
    def case(self):
        """Highest priority case-setter for new columns."""
        return self._case

    @case.setter
    def case(self, case):
        self._case = case
        if isinstance(case, str) and case is not None:
            case = getattr(str, case)
        
        cases = [str.lower, str.title, str.capitalize, str.upper, None]
        if case not in cases:
            raise ValueError(
                'Only string-class case types are allowed, or None. '
                'Choose one of [str.lower, str.title, str.capitalize, '
                'str.upper, None].'
            )
        
        # Update cases in plate attributes
        if case is not None:
            new_column_dict = {
                key: [case(col) for col in value]
                for key, value in self.column_dict.items()
            }
        
            # Update column dict
            self._column_dict = new_column_dict
            
            # Update casing in order, since that does not change
            columns = [(k1, case(k2)) for (k1, k2) in self.mi_df.columns]
            self._well = case(self._well)
            self.mi_df.columns = pd.MultiIndex.from_tuples(columns)
            self.mi_df.index.name = self._well
            
            # Reset dataframe
            self.df = self._from_midf()

    def _set_case(self, string):
        """For actually setting the case."""
        # If case is enforced with Plate.case
        if self.case is not None and string is not None:
            string = self.case(string)
        return string

### Zero padding of well
    @property
    def zero_padding(self):
        """Determines if well inputs are zero-padded."""
        return self._zero_padding

    @zero_padding.setter
    def zero_padding(self, padding):
        self._zero_padding = padding
        wells = self.df[self._well]

        # If no explicit input, assume False and switch conditionally
        if padding is None:
            self._zero_padding = False
            padded = [True for well in wells if _infer_padding(well)]

            if padded:
                self._zero_padding = True

        self.mi_df.index = wells.apply(pad, padded=self.zero_padding)
        self.df = self._from_midf()

### Value name
    @property
    def value_name(self):
        """Sets value name from `value_name`, or sets to 'value',
        unless obtained from the final column of a DataFrame.
        """
        return self._value_name

    @value_name.setter
    def value_name(self, value_name):
        self._value_name = value_name
        # Rearrange data
        values = self._values
        try:
            values.remove(value_name)
        except ValueError:
            pass
        values.append(value_name)
        self.values = values

### Primary descriptors
    @property
    def locations(self):
        """Each column in the data that specifies location information"""
        return self._locations

    @locations.setter
    def locations(self, locations):        
        # Ensure base locations are still there, and well is first
        base_locs = [self._well, self._row, self._col]
        for base_loc in base_locs:
            if base_loc not in locations:
                if base_loc == self._well:
                    locations = [self._well, *locations]
                else:
                    locations.append(base_loc)
        
        self._locations = locations
        
        # Remove if found in annotations
        ann_check = set(self.locations) & set(self.annotations)
        if ann_check:
            self._annotations = [ann for ann in self._annotations
                                 if ann not in ann_check]
        
        # Same for values
        val_check = set(self.locations) & set(self.values)
        if val_check:
            self._values = [val for val in self._values
                            if val not in val_check]
        
        # Reset column_dict to rearrange dataframe
        self.column_dict = {
            'locations': self.locations,
            'annotations': self.annotations,
            'values': self.values,
        }

    @property
    def annotations(self):
        """Each column in the data that specifies additional information"""
        return self._annotations

    @annotations.setter
    def annotations(self, annotations):
        self._annotations = annotations
        
        # Remove if found in locations
        loc_check = set(self.annotations) & set(self.locations)
        if loc_check:
            self._locations = [loc for loc in self._locations
                                 if loc not in loc_check]
        
        # Same for values
        val_check = set(self.annotations) & set(self.values)
        if val_check:
            self._values = [val for val in self._values
                            if val not in val_check]
        
        # Reset column dict to rearrange dataframe
        self.column_dict = {
            'locations': self.locations,
            'annotations': self.annotations,
            'values': self.values,
        }

    @property
    def values(self):
        """Each column in the data that specifies value information"""
        return self._values

    @values.setter
    def values(self, values):
        # Check that there is at least one value
        if not values:
            raise ValueError(
               'Plate must contain at least one value.'
            )
        
        self._values = values
        
        # Remove if found in locations
        loc_check = set(self.values) & set(self.locations)
        if loc_check:
            self._locations = [loc for loc in self._locations
                                 if loc not in loc_check]
        
        # Same for annotations
        ann_check = set(self.values) & set(self.annotations)
        if ann_check:
            self._annotations = [ann for ann in self._annotations
                                 if ann not in ann_check]
        
        # Reset column dict to rearrange dataframe
        self.column_dict = {
            'locations': self.locations,
            'annotations': self.annotations,
            'values': self.values,
        }

        # Update value name
        self._value_name = self.values[-1]

    @property
    def column_dict(self):
        return self._column_dict

    @column_dict.setter
    def column_dict(self, column_dict):
        self._column_dict = column_dict

        # Confirm values list still has members
        if not column_dict['values']:
            # If not, reset it and raise error
            old_values = list(self.mi_df.xs('values', axis=1).columns)
            column_dict['values'] = old_values
            for attr_list in (self.locations, self.annotations):
                for val in old_values:
                    if val in attr_list:
                        attr_list.remove(val)
            raise ValueError(
                'Plate must contain at least one value.'
            )
        # Ensure base locations are still there
        locs = column_dict['locations']
        base_locs = [self._well, self._row, self._col]
        for base_loc in base_locs:
            if base_loc not in locs:
                if base_loc == self._well:
                    locs = [self._well, *locs]
                else:
                    locs.append(base_loc)
      
        column_dict['locations'] = locs

        # Re-arrange DataFrame
        # Get current multi-index df dict
        mi_df = self.mi_df.reset_index(col_level=1, col_fill='locations')
        mi_dict = mi_df.to_dict()
       
        # Re-arrange columns
        new_dict = {}
        # Iterate through key pairs
        for key1, value in column_dict.items():
            for key2 in value:
                try:
                    # Grab data from same key pair
                    series = mi_dict[(key1, key2)]
                except KeyError:
                    # If attribute key has changed, grab data from df
                    series = self.df[key2]
                # Store in dictionary
                new_dict[(key1, key2)] = series
     
        # Assign as new mi_df and sort
        self.mi_df = pd.DataFrame(new_dict)
        self.mi_df = self.mi_df.sort_values(
            by=[('locations', loc) for loc in self.locations[1:]]
        ).reset_index(drop=True)
        self.mi_df = self.mi_df.set_index(('locations', self._well))
        self.mi_df.index.name = self._well

        # Reset dataframe
        self.df = self._from_midf()

        # Set each grouping
        self._locations = self.column_dict['locations']
        self._annotations = self.column_dict['annotations']
        self._values = self.column_dict['values']

###########################
# DataFrame construction
###########################
    def _generate_df(self):
        """Parses the column names in the initial DataFrame. Returns the
        strings found for 'well', and 'value_name'. Everything is
        assumed to work in this function (i.e, no checks performmed)
        since this only runs after check_inputs(Plate) is run.
        """
        self.df = self.data.copy()
        column_names = list(self.df.columns)

        # Deal with simple data first
        if column_names[-1] is None:
            if self.value_name is None:
                self._value_name = self._set_case('value')

            if self._case is None:
                # Check casing on value_name
                cases = _infer_casing(self.value_name)

                # Assign case for well string (self._well)
                if not cases:
                    cases.append('lower')
                case = getattr(str, cases[0])
                
                self._well = case('well')

            else:
                # If case is enforced, enforce it
                case = self._set_case
                self._value_name = case(self.value_name)
            
            # Apply to generated column names
            self._well = case('well')
            self._row = case('row')
            self._col = case('column')

            # Update data
            column_names[0] = self._well
            column_names[-1] = self.value_name
            self.df.columns = column_names

        # Now deal with any generic DataFrame
        if self.value_name is None:
            self._value_name = column_names[-1]
        else:
            # Move value_name to end of dataframe
            vals = self.df[self.value_name].values
            del self.df[self.value_name]
            self.df[self.value_name] = vals

        # Determine well string
        self._well = [col for col in column_names if col.lower() == 'well'][0]

        # Move well to front
        vals = self.df[self._well].values
        del self.df[self._well]
        self.df.insert(0, self._well, vals)

        # Generate row and column info
        row = [col for col in self.df.columns if col.lower() == 'row']
        col = [col for col in self.df.columns if col.lower() == 'column']

        if not row:
            # Keep earlier string, if it was made
            try:
                row.append(self._row)
            except AttributeError:
                row.append('row')
        if not col:
            # Keep earlier string, if it was made
            try:
                col.append(self._col)
            except AttributeError:
                col.append('column')
        self._row = self._set_case(row[0])
        self._col = self._set_case(col[0])
        try:
            del self.df[self._row]
            del self.df[self._col]
        except KeyError:
            pass

        # Generate row and column data
        rows = [well[0] for well in self.df[self._well]]
        cols = [int(well[1:]) for well in self.df[self._well]]

        # Add in correct order
        self.df.insert(1, self._col, cols)
        self.df.insert(1, self._row, rows)

        return self.df, self._well, self._row, self._col

    def _generate_mi_df(self):
        """Sets up a multi-index DataFrame, for easier arraying and
        slicing of data."""
        # Well as index
        mi_df = self.df.copy().set_index(self._well)

        # Set up multi-index columns
        tuples = []
        for key, value in self.column_dict.items():
            tuples += [(key, val) for val in value
                       if val != self._well]

        # Create the multi-index
        mi_cols = pd.MultiIndex.from_tuples(tuples)
        
        # Rename columns and move data appropriately
        mi_df.columns = mi_cols
        for col in mi_df.columns:
            mi_df[col] = self.df[col[:][1]].values

        return mi_df

    def _from_midf(self):
        """Create df from mi df"""
        self.df = self.mi_df.copy()
        self.df = self.df.reset_index(col_level=1)
        self.df.columns = self.df.columns.droplevel(0)
        return self.df

###########################
# Plate methods
###########################

    def set_as_location(self, loc, idx=-1):
        """Sets a column as a location"""
        plate = self.copy()

        # Check column
        check_df_col(plate.df, loc, 'location')

        # Allow idx arg to rearrange instead of fail by removing loc if present
        if loc in plate.locations:
            plate.locations.remove(loc)

        # Set up new locations attribute
        locs = plate._locations
        
        if idx == -1:
            locs.append(loc)
        else:
            idx += 1
            locs = [
                *locs[:idx],
                loc,
                *locs[idx:]
            ]

        plate.locations = locs

        return plate

    def set_as_value(self, new_values=None, value_name=None):
        """Sets columns in the Plate DataFrame as values.
        
        Parameters:
        -----------
        new_values: str or list of str
            A list of column names to be set as values.
        value_name: str
            Which value should be set as the main (-1 index) value.
        """
        plate = self.copy()

        if set(plate.values) & set(new_values):
            raise ValueError(
                'Cannot set current values as new values.'
            )
        # Set as list
        if new_values is not None:
            if not isinstance(new_values, list):
                new_values = [new_values]
                values = new_values + plate.values

        # Update plate.value_name
        if value_name is not None:
            values.remove(value_name)
            values.append(value_name)
        else:
            value_name = values[-1]

        # Move main value to end of vals list
        if value_name not in [*plate.values, *values]:
            raise ValueError(
                f"'{value_name}' not found in list of values. Pick one of "\
                f"{plate.values} or pass new_values='{value_name}'."
            )
      
        plate.values = values

        return plate
      
    def annotate_wells(self, annotations):
        """Takes either a nested dictionary or standardized excel
        spreadsheet (see ninetysix/templates) to assign new columns
        in the output DataFrame that provide additional information
        about the contents or conditions of each well in the Plate.

        Parameters:
        -----------
        annotations: nested dictionary or excel sheet
            A mapping that assigns conditions to wells. For a nested
            dictionary, the outer key(s) will be the name of the condition
            which results in a new DataFrame column for each condition
            and each inner dictionary should contain key-value pairs
            where the keys either a well/regex-like well (e.g., A1 or
            [A-C]1) or a default ('default', 'standard', 'else', 'other')
            and the values are the value of the specific well
            for that condition. For an excel sheet, it should be from
            a template (https://github.com/palmhjell/ninetysix/templates).
        """
        plate = self.copy()
        
        # Check that everything passes, return type.
        annotation_type = check_annotations(self, annotations)

        if annotation_type == dict:
            # Unpack dictionary and assign
            for column in annotations.keys():

                working_annotations = well_regex(annotations[column],
                                                 padded=plate.zero_padding)

                # Check for default
                default = None
                acceptable_kwargs = ('default', 'standard', 'else', 'other')
                for key in working_annotations.keys():
                    if key.lower() in acceptable_kwargs:
                        default = working_annotations[key]

                # Make new columns
                key = ('annotations', column)
                plate.mi_df[key] = plate.mi_df.index.map(working_annotations.get)
                plate.mi_df[key] = plate.mi_df[key].replace({None: default})

            new_annotations = [col for col in annotations.keys()
                               if col not in plate.annotations]
            
            plate.annotations = [*plate.annotations, *new_annotations]

        elif annotation_type == 'excel':

            # Read the mapping spreadsheet
            df_map = pd.read_excel(annotations, sheet_name=0, index_col=[1, 0])

            # Initialize lists (only 96-well right now)
            rows = list('ABCDEFGH')
            df_list = []

            # Iterate through each row index
            for row in rows:

                # Get tidy data from each row
                working_df = df_map.loc[row].T

                # Add which row it is
                working_df['row'] = row

                # Add to list for later concat
                df_list.append(working_df)

            # Create a new DataFrame
            df_map = pd.concat(df_list, sort=True)

            # Get columns from index
            df_map.index.name = 'column'
            df_map = df_map.reset_index()

            # Add Well column
            wells = [f'{row}{column}'
                     for (row, column) in zip(df_map['row'], df_map['column'])]
            df_map['well'] = [pad(well, plate.zero_padding) for well in wells]

            # Drop columns that are *entirely* nans
            df_map = df_map.dropna(axis='columns', how='all')

            # Switch to None
            df_map = df_map.replace({np.nan: None})

            # Standardize case and merge
            df_map.columns = [plate._set_case(col)
                             for col in df_map.columns]

            mergers = [plate._set_case(col)
                      for col in ['well', 'column', 'row']]

            plate.df = plate.df.merge(df_map, on=mergers)

            # Update annotations
            new_annotations = [col for col in df_map
                               if col not in mergers
                               and col not in plate.annotations]
           
            plate.annotations = [*plate.annotations, *new_annotations]

        return plate


    def normalize(
        self,
        value=None,
        to=None,
        zero=None,
        groupby=None,
        # devs=False,
        update_value=None,
        prefix='normalized_',
    ):
        """Normalizes the value column to give the max a value of 1,
        returning a new column named '[prefix][value]', e.g.,
        'normalized_value. Accepts different value kwargs and can
        explicitly scale from 0 to 1 'zero=True'. Alternatively, scales
        relative to a specific assignment of a condition column, i.e. to 'Standard' within the condition 'Controls' can be set to a value
        of 1 via the kwarg `to='Controls=Standard'`. Additionally can
        assign a lower 0 value in the same way.

        Parameters:
        -----------
        value: str or list of str, default None
            Name of column to normalize, default being value_name.
        to: str, default None
            Which group to set as the normal (1) value. If None, the max
            value for each column in `value` is set to 1 and all other
            values are scaled to this. If a condition is passed, e.g.,
            `to='Controls=Standard'`, the mean of the wells labeled
            'Standard' in the column 'Controls' will be set to 1 for
            each value column.
        zero: str or bool, default None
            Whether or not to set the lowest value as an explicit zero
            (`zero=True`) or just to scale all values as is (`zero=
            False`), or to set the mean of a specific group as the zero
            value as in the `to` kwarg, e.g., `zero='Controls=Negative'`
            would set the  wells labeled 'Negative' in the column
            'Controls' to zero (this is useful when your 'zero'
            assay condition does not actually give values of zero).
        groupby: str, default None
            Whether to split the normalization across many groups in the
            column (or columns) specified in groupby, e.g., for each
            value in the column 'Plate'.
        update_value: str or bool, default None
            Whether or not (or what) to set as the new Plate value_name.
            If True and `value` is not a list, sets `value` to right-
            most column. If a string, sets that normalized value to the
            new `value_name`. If None or False, just adds a new value
            and leaves the old `value_name`. 
        prefix: str, default 'normalized_'
            What to add to the new column name to indicate that it has
            been normalized.
        """
        plate = self.copy()

        # Determine how to update the value
        # TODO: does this need a warning for True when type(value) == list?
        if isinstance(value, list) and isinstance(update_value, str):
            if update_value not in value:
                raise ValueError(
                    f"Given update value '{update_value}' not found in list of values to be normalized."
                )

        # Get value list ready
        if value is None:
            value = plate.value_name
        if not isinstance(value, list):
            values = [value]
        else:
            values = value

        # Set up groups
        if not isinstance(groupby, (tuple, list)):
            groupby = [groupby]

        for group in groupby:
            check_df_col(plate.df, group, name='groupby')

        # Group or create fake groupby object
        unique_dfs = (plate.df.groupby(groupby) if groupby != [None]
                      else [(None, plate.df.copy())])

        df_list = []
        # Iterate through each dataframe group
        for name, sub_df in unique_dfs:
            
            sub_df = sub_df.copy()
            
            # Iterate through each value
            for value in values:
                check_df_col(sub_df, value, name='value')

                norm_string = f'{prefix}{value}'

                # Set the zero val
                if zero == True:
                    zero_val = sub_df[value].min()
                elif isinstance(zero, str):
                    split = zero.split('=')
                    if len(split) != 2:
                        raise ValueError(
                            f"'zero' value specified incorrectly, must be of the form 'column_name=value_name'."
                        )

                    col, val = split

                    # Check that col in columns
                    check_df_col(sub_df, col, name='column')

                    # Check that the given value is found in the column
                    if val not in [str(i) for i in sub_df[col]]:
                        raise ValueError(
                            f"The value '{val}' is not a value in the column '{col}'."
                        )
                    subset = [val == str(i) for i in sub_df[col]]
                    zero_val = sub_df[subset][value].mean()
                elif zero is None or zero == False:
                    zero_val = 0
                else:
                    raise TypeError(
                        f"Type of 'zero' argument is incorrectly specified. Must be bool or string.")

                sub_df[norm_string] = sub_df[value] - zero_val

                # Set the one val
                if to is None:
                    one_val = sub_df[norm_string].max()
                elif isinstance(to, str):
                    split = to.split('=')
                    if len(split) != 2:
                        raise ValueError(
                            f"'to' value specified incorrectly, must be of the form 'column_name=value_name'."
                        )
                    
                    col, val = split
                    
                    # Check that col in columns
                    check_df_col(sub_df, col, name='column')

                    # Check that the given value is found in the column
                    if val not in [str(i) for i in sub_df[col]]:
                        raise ValueError(
                            f"The value '{val}' is not a value in the column '{col}'."
                        )
                    subset = [val == str(i) for i in sub_df[col]]
                    one_val = sub_df[subset][norm_string].mean()
                else:
                    raise TypeError(f"Type of 'to' argument is incorrectly specified. Must be bool or string.")

                sub_df[norm_string] = sub_df[norm_string] / one_val

            # After adding normalized values, store in df_list
            df_list.append(sub_df)

        # Merge dataframes from groupby
        mergers = [plate._well,
                   *[elem for elem in groupby if elem is not None]]
        full_df = plate.df[mergers].merge(
            pd.concat(df_list), on=mergers
        )
        
        # Add each new value
        for value in values:
            norm_string = f'{prefix}{value}'
            plate.mi_df[('values', norm_string)] = full_df[norm_string].values
            
            # Update plate.values list
            if norm_string not in plate.values:
                val_idx = plate.values.index(value)

                # Add norm value to the left of value
                plate.values = [
                    *plate.values[:val_idx],
                    norm_string,
                    *plate.values[val_idx:]
                ]
        
        # Clean up
        plate.df = plate._from_midf()
        if update_value:
            if isinstance(update_value, str):
                update_string = f'{prefix}{update_value}'
            else:
                update_string = norm_string
            
            plate.value_name = update_string
            plate.values.remove(update_string)
            plate.values = [*plate.values, update_string]

        return plate
  



"""
Visualization functions for Plate objects or any generic DataFrame.
"""

import numpy as np
import pandas as pd

import holoviews as hv
hv.extension('bokeh')
import bokeh.palettes

from .util import check_df_col

class Colors():
    """Mild color palette."""
    blue = '#3E94FA'
    green = '#8DED81'
    orange = '#E18409'
    gray = '#DEDEDE'


def _get_ordered_locs(df):
    """Returns location information in 'well', 'row', 'column' order."""
    well = [val for val in df.columns if val.lower() == 'well']
    row = [val for val in df.columns if val.lower() == 'row']
    column = [val for val in df.columns if val.lower() == 'column']

    for loc in (well, row, column):
        if not loc:
            raise ValueError(
                f'Could not find column for "{loc}"'
            )
        if len(loc) > 1:
            raise ValueError(
                f'Too many columns found: "{loc}"'
            )
    
    return well[0], row[0], column[0]


def _parse_data_obj(obj):
    """Gets necessary information for 96-well-based plots"""
    if isinstance(obj, pd.DataFrame):
        df = obj.copy()
        value_name = df.columns[-1]

        # Determine case
        lowers = [True for val in df.columns if val.lower() == val]
        caps = [True for val in df.columns if val.capitalize() == val]
        case = str.lower if lowers > caps else str.capitalize

    # Assume Plate
    else:
        try:
            df = obj.df.copy()
            value_name = obj.value_name
            case = obj._set_case
        except AttributeError:
            raise ValueError(
                'Data format not recognized, only pandas.DataFrame and '\
                'ninetysix.Plate objects accepted.'
            )

    # Get well, row, and column values
    try:
        well, row, col = _get_ordered_locs(df)
        locs = (well, row, col)
    except ValueError:
        locs = None

    return df, locs, value_name, case


def _center_colormap(data, cmap_center):
    """Stretch a color map so that its center is at `cmap_center`."""
    if not isinstance(cmap_center, (float, int)):
        raise TypeError(
            'cmap_center argument must be float or int.'
        )
    if not min(data) < cmap_center < max(data):
        raise ValueError('Must have min(data) < cmap_center < max(data).')

    dist = max(max(data) - cmap_center, cmap_center - min(data))
    dist += dist / 100

    return list(np.linspace(cmap_center-dist, cmap_center+dist, 257))


def aggregate_replicates(df, variable, value, grouping):
    """Checks for the presence of replicates in the values of a dataset,
    given some experimental conditions. Returns True if the standard
    deviation of the values of each group (if more than one exists) is
    greater than 0, indicating that replicates were performed under the
    given criteria.
    
    Parameters:
    -----------
    df: pandas DataFrame in tidy format
        The dataset to be checked for replicates.
    variable: string
        Name of column of data frame for the independent variable,
        indicating a specific experimental condition.
    value: string
        Name of column of data frame for the dependent variable,
        indicating an experimental observation.
    grouping: list of strings
        Column name or list of column names that indicates how the
        data set should be split.
    
    Returns:
    --------
    replicates: bool
        True if replicates are present.
    df_agg: pd.DataFrame
        The DataFrame containing averaged 'variable' values, if
        replicates is True. Otherwise returns the original DataFrame.
        The aggregated data is in a new column
    """
    # Unpack the experimental conditions into a single list of arguments
    if not isinstance(grouping, (list, tuple)):
        grouping = [grouping]
    args = [elem for elem in (variable, *grouping) if elem is not None]

    # Get stdev of argument groups
    grouped = df.groupby(args)[value]
    group_stdevs = grouped.std().reset_index()
    group_stdev = group_stdevs[value].mean()

    # Determine if there are replicates (mean > 0)
    replicates = bool(group_stdev > 0)

    # Average the values and return
    if replicates:
        df_mean = grouped.mean().reset_index()
        df_mean.columns = list(df_mean.columns[:-1]) + ['mean_' + str(value)]
        df = df.merge(df_mean)

    return replicates, df


def plot_scatter(
    obj,
    variable=None,
    value_name=None,
    color=None,
    layering=None,
    layering_order=None,
    cmap='CategoryN',
    ranked=False,
    groupby=None,
    layout=False,
    n_cols=2,
    height=350,
    width=450,
    plot_opts=None,
):
    """Create a well-activity scatter plot.

    Parameters:
    -----------
    obj: ns.Plate or pd.DataFrame objectect
        Must contain a DataFrame with columns labeled well, row, column,
        (case-insensitive) and the final column as the value (can be
        overwritten, see `value_name` kwarg).
    variable: string, default None
        If None, `obj` must contain a well column. Otherwise specifies
        the variable on the x-axis of the scatter chart.
    value_name: string, default None
        Which column in the data contains the y-axis data. Defaults to
        ns.Plate.value_name or pd.DataFrame.columns[-1].
    color: string, default None
        Either the column that should be used to color the data or the
        color all data points should be (e.g., 'green' or '#00FF00').
        If there is a column in the data that corresponds to the `color`
        input, it will be used first (before rgb/hex colors).
    layering: string, default None
        Which column in the data should be used to order the layering
        of the data. This kwarg is useful when you want certain points
        (e.g., control wells) to not be covered by other points. See 
        `layering_order` kwarg to control the specific ordering (and
        coloring) of the values in this column.
    layering_order: 'ascending', 'descending', or a list or dict.
        Layering of the data, from front to back, using the values in
        the column specified in `layering`. If providing a dictionary,
        keys are provided as the layering (in order) and values are the
        respective colors of each group.
    cmap: string, list, or dict, default 'CategoryN'
        How to color the groups of data found in `color`. Default of 
        'CategoryN' switches between 'Category10' and 'Category20'
        depending on the number of groups. If a different string, must
        be an acceptable cmap for holoviews. Otherwise, can be a list of
        colors (which matches the order of `layering`, if not None and
        if `layering_order` is not a dict), or the same dict as
        `layering_order`.
    ranked: bool, default False
        Whether to rank the data (as a retention-of-function-like curve)
        or not (as a well-sorted scatter plot). Passing False is good to
        check signal-to-noise and edge effects in the plate.
    groupby: string or list, default None
        One or more columns in the data used to split the plot into
        multiple plots, usually a location (like plate number) or
        annotation.
    layout: bool, default False
        Whether or not to lay out all of the data (True) or keep it as
        a slider-based panel (False), if `groupby` was not None.
    n_cols: int, default 2
        Number of columns of laid-out plots if `layout` is True.
    height: int, default 350
        Height of the plot.
    width: int, default 450
        Width of the plot.
    plot_opts: dict, default None
        Holoviews-specific opts to pass to the chart. See the holoviews
        docs. Overwrites the base plot opts if the same opt is passed.

    Returns:
    --------
    p: Holoviews plot
    """
    if plot_opts is None:
        plot_opts = {}
    # Get data and metadata
    df, locs, auto_value, case = _parse_data_obj(obj)
    if value_name is not None:
        check_df_col(df, value_name, 'value_name')
    else:
        value_name = auto_value
    
    try:
        check_df_col(df, color, 'color')
    except ValueError:
        # Assuming color is not column name but actual color; test
        try:
            hv.Scatter(df).opts(color=color)
        except ValueError:
            raise ValueError(
                f'Unusable color kwarg. {color} not an acceptable option.'
            )

    if color is not None:
        plot_opts['color'] = color
    
    check_df_col(df, variable, 'variable')
    check_df_col(df, layering, 'layering')
    
    if not isinstance(groupby, list):
        groupby = [groupby]
    for group in groupby:
        check_df_col(df, group, 'groupby')

    if case('rank') not in df.columns:
        if groupby == [None]:
            df[case('rank')] = df[value_name].rank(ascending=False, method='first')
        else:
            grouped = df.groupby(groupby)[value_name]
            df[case('rank')] = grouped.rank(ascending=False, method='first')
    
    secondary_sort = []
    kdims = []
    if ranked:
        secondary_sort.append(case('rank'))
        kdims.append(case('rank'))
    elif variable is None:
        if locs is None:
            raise ValueError(
                'Data object does not contain column related to well. This is '
                'required if a variable is not specified and ranked is False.'
            )
        plot_opts['xticks'] = 0
        # Order by well (A1-H12 ordering)
        secondary_sort += [locs[1], locs[2]]
        # kdims = well
        kdims = locs[0]
    else:
        kdims = variable
        secondary_sort.append(variable)


    
    df = df.sort_values(by=secondary_sort, ascending=True)

    # Layering
    if layering is not None:
        if layering_order == 'ascending':
            df = df.sort_values(by=[layering, *secondary_sort], ascending=True)
        if layering_order == 'descending':
            df = df.sort_values(by=[layering, *secondary_sort], ascending=False)
        if isinstance(layering_order, list):
            df[layering] = pd.Categorical(df[layering], categories=layering_order)
            # Reverse ordering to line up
            cmap=list(reversed(cmap))
        if isinstance(layering_order, dict):
            df[layering] = pd.Categorical(
                df[layering],
                categories=layering_order.keys()
            )
            # Reverse ordering to line up with layering
            cmap = list(reversed(list(layering_order.values())))
        if isinstance(cmap, dict):
            df[layering] = pd.Categorical(
                df[layering],
                categories=cmap.keys()
            )
            # Reverse ordering to line up with layering
            cmap = list(reversed(list(cmap.values())))

        df = df.sort_values(by=[layering, *secondary_sort], ascending=not ranked)

    elif color is not None:
        if isinstance(cmap, dict):
            df[color] = pd.Categorical(
                df[color],
                categories=cmap.keys()
            )
            # Reverse ordering to line up with layering
            if ranked:
                cmap = list(reversed(list(cmap.values())))
            else:
                cmap = list(cmap.values())

            df = df.sort_values(by=[color, *secondary_sort], ascending=not ranked)

    vdims = [value_name] + [val for val in df.columns if val != value_name]

    # Auto-colomapping
    if cmap == 'CategoryN':
        cmap = 'Category10'
        try:
            if len(df[color].unique()) > 10:
                cmap = 'Category20'
        except KeyError:
            pass

    # Set up options; entries can be overwritten by plot_opts
    base_opts = dict(
        height=height,
        width=width,
        size=10,
        fill_alpha=1,
        alpha=0.75,
        cmap=cmap,
        fontsize=dict(title=9),
        line_width=2,
        line_color='black',
        padding=0.05,
        toolbar='above',
        tools=['hover'],
        **plot_opts,
    )

    p = hv.Scatter(
        df,
        kdims,
        vdims
    )
    
    if groupby != [None]:
        p = p.groupby(groupby)
    
    p = p.opts(**base_opts)

    if layout and grouped != [None]:
        p = p.layout().cols(n_cols)

    return p


def plot_rof(*args, **kwargs):
    """Create a retention of function curve, or a rank-ordered scatter
    plot. See `plot_scatter` for more details.

    Parameters:
    -----------
    object: ns.Plate or pd.DataFrame objectect
        Must contain a DataFrame with columns labeled well, row, column,
        (case-insensitive) and the final column as the label (can be
        overwritten, see `value_name` kwarg).
    value_name: string, default None
        Which column in the data contains the y-axis data. Defaults to
        ns.Plate.value_name or pd.DataFrame.columns[-1].
    color: string, default None
        Either the column that should be used to color the data or the
        color all data points should be (e.g., 'green' or '#00FF00').
        If there is a column in the data that corresponds to the `color`
        input, it will be used first (before rgb/hex colors).
    layering: string, default None
        Which column in the data should be used to order the layering
        of the data. This kwarg is useful when you want certain points
        (e.g., control wells) to not be covered by other points. See 
        `layering_order` kwarg to control the specific ordering (and
        coloring) of the values in this column.
    layering_order: 'ascending', 'descending', or a list or dict.
        Layering of the data, from front to back, using the values in
        the column specified in `layering`. If providing a dictionary,
        keys are provided as the layering (in order) and values are the
        respective colors of each group.
    cmap: string, list, or dict, default 'CategoryN'
        How to color the groups of data found in `color`. Default of 
        'CategoryN' switches between 'Category10' and 'Category20'
        depending on the number of groups. If a different string, must
        be an acceptable cmap for holoviews. Otherwise, can be a list of
        colors (which matches the order of `layering`, if not None and
        if `layering_order` is not a dict), or the same dict as
        `layering_order`.
    ranked: bool, default True
        Whether to rank the data (as a retention-of-function-like curve)
        or not (as a well-sorted scatter plot). Passing False is good to
        check signal-to-noise and edge effects in the plate.
    groupby: string or list, default None
        One or more columns in the data used to split the plot into
        multiple plots, usually a location (like plate number) or
        annotation.
    layout: bool, default False
        Whether or not to lay out all of the data (True) or keep it as
        a slider-based panel (False), if `groupby` was not None.
    n_cols: int, default 2
        Number of columns of laid-out plots if `layout` is True.
    height: int, default 350
        Height of the plot.
    width: int, default 450
        Width of the plot.
    plot_opts: dict, default None
        Holoviews-specific opts to pass to the chart. See the holoviews
        docs. Overwrites the base plot opts if the same opt is passed.

    Returns:
    --------
    p: Holoviews plot
    """
    if 'ranked' not in kwargs:
        kwargs['ranked'] = True
    return plot_scatter(*args, **kwargs)


def plot_hm(
    obj,
    value_name=None,
    outline=None,
    exclude_major=False,
    groupby=None,
    layout=False,
    n_cols=2,
    hm_cmap='RdBu_r',
    cmap_center=None,
    outline_cmap='CategoryN',
    legend=True,
    height=None,
    hm_opts=None,
    colorbar_opts=None,
    outline_opts=None,
):
    """Plot a row-by-column heatmap.

    Parameters:
    -----------
    obj: ns.Plate or pd.DataFrame object
        Must contain a DataFrame with columns labeled well, row, column,
        (case-insensitive) and the final column as the label (can be
        overwritten, see `value_name` kwarg).
    value_name: string, default None
        Which column in the data contains the heatmap coloring data.
        Defaults to ns.Plate.value_name or pd.DataFrame.columns[-1].
    outline: string, default None
        Which column in the data contains groups to be outlined for easy
        visualization.
    exclude_major: bool, default False
        Whether or not to exclude the major group from the outline to
        reduce the noise in the plot (i.e., excluding the experimental
        wells and only showing the controls).
    groupby: string or list, default None
        One or more columns in the data used to split the plot into
        multiple plots, usually a location (like plate number) or
        annotation.
    layout: bool, default False
        Whether or not to lay out all of the data (True) or keep it as
        a slider-based panel (False), if `groupby` was not None.
    n_cols: int, default 2
        Number of columns of laid-out plots if `layout` is True.
    hm_cmap: string or continuous colormap list, default 'RdBu_r'.
        How to assign values to colors on the heatmap. If a different
        string, must be an acceptable cmap for holoviews. Otherwise, can
        be a list of colors from a continuous colormap.
    cmap_center: float or int, default None
        The center of the heatmap's colormap. Useful for setting the
        center of a diverging colormap to an important value (like a
        control).
    outline_cmap: string, list, or dict, default 'CategoryN'
        How to color the groups of data found in `outline`. Default of 
        'CategoryN' switches between 'Category10' and 'Category20'
        depending on the number of groups. If a different string, must
        be an acceptable cmap for holoviews. Otherwise, can be a list of
        colors. If providing a dictionary, keys are provided as the
        groups (in order) and values are the respective colors of each
        group.
    legend: bool, default False
        Whether to show the outline legend.
    height: int, default None
        Height of the chart. Defaults to each row having a height of 50.
        Width is then determined automatically to keep wells square. If
        non-square wells are needed, width can be overwritten in
        `hm_opts` kwarg.
    hm_opts: dict, default None
        Holoviews-specific opts to pass to the heatmap. See the
        holoviews docs. Overwrites the base opts if the same opt is
        passed.
    colorbar_opts: dict, default None
        Holoviews-specific opts to pass to the heatmap's colorbar. See
        the holoviews docs. Overwrites the base opts if the same opt is
        passed.
    outline_opts: dict, default None
        Holoviews-specific opts to pass to the outline chart. See the
        holoviews docs. Overwrites the base opts if the same opt is
        passed.

    Returns:
    --------
    p: Holoviews plot
    """
    if hm_opts is None:
        hm_opts = {}
    if colorbar_opts is None:
        colorbar_opts = {}
    if outline_opts is None:
        outline_opts = {}
    # Get data and metadata
    df, locs, auto_value, case = _parse_data_obj(obj)
    if locs is None:
        raise ValueError(
            'DataFrame is missing well, row, and/or column information. '
            'This is necessary for this visualization.'
        )
    if value_name is not None:
        check_df_col(df, value_name, 'value_name')
    else:
        value_name = auto_value
    if outline is not None:
        check_df_col(df, outline, 'outline')
    if not isinstance(groupby, list):
        groupby = [groupby]
    if groupby != [None]:
        for group in groupby:
            check_df_col(df, group, 'groupby')

    # Non-int column
    df[locs[2]] = df[locs[2]].astype(str)

    kdims = [locs[2], locs[1]]
    undims = (value_name, *kdims)
    vdims = [value_name] + [val for val in df.columns if val not in undims]

    # Adjust heights
    hm_rows = len(df[locs[1]].unique())
    hm_cols = len(df[locs[2]].unique())
    if height is None:
        height = int(50 * hm_rows)
    width = height * hm_cols // hm_rows

    # Center colorbar
    if cmap_center is None:
        # Default to middle of data range (normal behavior)
        values = df[value_name]
        cmap_center = min(values) + (max(values) - min(values))/2

    color_levels = _center_colormap(df[value_name].dropna(), cmap_center)

    # Set the plot options; can be overwritten
    base_opts = dict(
        cmap=hm_cmap,
        xaxis='top',
        frame_height=height,
        frame_width=width,
        colorbar=True,
        color_levels=color_levels,
        invert_yaxis=True,
        tools=['hover'],
        line_width=2,
        **hm_opts,
    )

    base_colorbar_opts = dict(
        title=value_name,
        title_text_align='left',
        height=int(0.9*height),
        padding=40,
        location=(-30, -20),
        bar_line_color=None,
        major_tick_line_color=None,
        label_standoff=5,
        background_fill_color=None,
        **colorbar_opts,
    )

    base_opts['colorbar_opts'] = base_colorbar_opts

    # Make the plot
    p = hv.HeatMap(
        df,
        kdims,
        vdims
    )
    
    if groupby != [None]:
        p = p.groupby(groupby)
        
    p = p.opts(**base_opts)

    # Add outlines
    major = None
    if outline is not None:
        if exclude_major:
            major = df[outline].value_counts().index[0]
            df_outline = df[df[outline] != major].copy()
        else:
            df_outline = df.copy()

        # Need string for legend to show; categorical stuff
        df_outline[outline] = df_outline[outline].astype(str)

        # Auto-colomapping
        if outline_cmap == 'CategoryN':
            if len(df_outline[outline].unique()) <= 10:
                outline_cmap = 'Category10'
            else:
                outline_cmap = 'Category20'
        
        if isinstance(outline_cmap, dict):
            # Remove major from dictionary if it exists
            outline_cmap = outline_cmap.copy()
            outline_cmap.pop(major, None)

            df_outline[outline] = pd.Categorical(
                df_outline[outline],
                categories=outline_cmap.keys()
            )
            df_outline = df_outline.sort_values(by=outline)
            # Reverse ordering to line up with order
            outline_cmap = list(outline_cmap.values())

        # Set up size of the outline boxes
        box_size = height // hm_rows - (height // hm_rows)*0.07

        boxes_opts = dict(
            color=outline,
            cmap=outline_cmap,
            show_legend=legend,
            marker='square',
            size=box_size,
            line_width=0.08*box_size,
            fill_alpha=0,
            legend_position='top',
            **outline_opts,
        )
        
        box_vdims = [val for val in (outline, *groupby) if val is not None]
        boxes = hv.Points(
            df_outline,
            kdims,
            box_vdims
        )
        
        if groupby != [None]:
            boxes = boxes.groupby(groupby)
            
        boxes = boxes.opts(**boxes_opts)

        p = p*boxes

        if layout and groupby != [None]:
            p = p.layout().cols(n_cols)

    return p


def plot_bar(
    obj,
    variable,
    value_name=None,
    groupby=None,
    color=None,
    sort=None,
    cmap='CategoryN',
    show_points=True,
    layout=False,
    n_cols=2,
    legend=False,
    xrotation=0,
    jitter=0,
    height=350,
    width=300,
):
    """Converts a tidy DataFrame a bar plot, taking care to show all
    the data. Bars are given as the average of each grouping of the
    variable (which can be further split by the split argument), and
    the actual data points are overlaid on top.
    
    Parameters:
    -----------
    obj: ns.Plate or pd.DataFrame object
        If DataFrame and `value = None`, the final column is used as the
        value.
    variable: str
        Column in DataFrame representing the variable, plotted on
        the x-axis.
    value_name: str, default None
        Column in DataFrame/Plate representing the quantitative value,
        plotted on the y-axis.
    groupby: str, default None
        The names of one or more columns that further specify the
        way the data is grouped.
    sort: str, default NNone.
        Which column is used to determine the sorting of the data.
        Defaults to None, and will sort by the condition column
        (alphabetical) if present, otherwise variable.
    cmap: str, default 'CategoryN'
        The colormap to use. Any Holoviews/Bokeh colormap is fine.
        Defaults to Category10 or Category20 depending on number of
        unique plot elements.
    show_points: bool, default True
        Shows all the data points. I don't even know why this is an
        argument. Default will show points if there are multiple
        replicates. Unless you have a really good reason, don't
        change this.
    layout: bool, default False
        Whether or not to lay out all of the data (True) or keep it as
        a slider-based panel (False), if `groupby` was not None.
    n_cols: int, default 2
        Number of columns of laid-out plots if `layout` is True.
    legend: bool/str, default False
        First controls whether or not the legend is shown, then its
        position. Defaults to False, though 'top' would be a good
        option, or 'top_left' if using split.
    xrotation: int, default 0
        Rotation of the x-labels in degrees.
    height: int, default 350
        The height of the chart.
    width: int, default 300
        The width of the chart.

    Returns:
    --------
    chart: Holoviews chart
    """
    # Get data and metadata
    df, locs, auto_value, case = _parse_data_obj(obj)
    if value_name is not None:
        check_df_col(df, value_name, 'value_name')
    else:
        value_name = auto_value
    
    # Check columns
    check_df_col(df, variable, 'variable')
    check_df_col(df, color, 'color')
    check_df_col(df, sort, 'sort')

    if not isinstance(groupby, (list, tuple)):
        groupby = [groupby]
    for group in groupby:
        check_df_col(df, group, 'groupby')

    # Auto-colomapping
    if cmap == 'CategoryN':
        if len(df[variable].unique()) <= 10:
            cmap = 'Category10'
        else:
            cmap = 'Category20'
    
    replicates, df = aggregate_replicates(df, variable, value_name, groupby)
    if f'mean_{value_name}' not in df.columns:
        df[f'mean_{value_name}'] = df[value_name]

    # Sort
    if sort is not None:
        check_df_col(df, sort, name='sort')
        df = df.sort_values(by=sort).reset_index(drop=True)
    
    # Check jitter value
    if not isinstance(jitter, (float, int)) or isinstance(jitter, bool):
        raise ValueError('Jitter must be an explicit numeric value.')
    if jitter > 1:
        # Paternalism (shout out to seaborn)
        raise ValueError('No. Jitter > 1 is misleading.')

    # Encode color
    if color is None:
        color = variable

    # If color is the value column, make it the mean
    if color == value_name:
        color = f'mean_{value_name}'

    # Pull out available encodings (column names)
    encodings = [*list(df.columns)]

    # Make bar chart
    bars = hv.Bars(
        df,
        variable,
        [(f'mean_{value_name}', value_name), *encodings],
    )

    # Determine single-point entries
    args = [elem for elem in [variable] + groupby if elem is not None]
    counts = (df.groupby(args).count() ==
                1).reset_index()[[variable, value_name]]
    counts.columns = [variable, 'counts']

    # Get list of singlets to drop from plotting df
    singlets = counts[counts['counts']][variable].to_list()

    # Make scatter chart
    points = hv.Scatter(
        df[~df[variable].isin(singlets)],
        variable,
        [value_name, *encodings],
    )

    # Make the split
    if groupby != [None]:
        bars = bars.groupby(groupby)
        points = points.groupby(groupby)

    # Output chart as only bars, or bars and points
    if show_points is None:
        if replicates:
            chart = bars*points
        else:
            chart = bars
    elif show_points:
        chart = bars*points
    else:
        chart = bars

    # Set options (this is probably horribly inefficient right now)
    base_opts = dict(
        height=height,
        width=width,
        ylim=(0, 1.1*np.max(df[value_name])),
        xrotation=xrotation,
        color=color,
        cmap=cmap,
        show_legend=legend,
    )

    bar_opts = base_opts
    scat_opts = dict(
        size=6,
        fill_alpha=0.2,
        color='black',
        jitter=jitter,
    )

    # Add to chart
    chart.opts({
        'Bars': bar_opts,
        'Scatter': scat_opts,
    })

    if layout and groupby != [None]:
        chart = chart.layout().cols(n_cols)

    return chart


def plot_curve(
    obj,
    variable,
    value_name=None,
    condition=None,
    groupby=None,
    sort=None,
    cmap='CategoryN',
    show_all=False,
    show_points=None,
    legend=False,
    height=350,
    width=500,
):
    """Converts a tidy DataFrame containing timecourse-like data
    into a plot, taking care to show all the data. A line is
    computed as the average of each set of points (grouped by the
    condition and groupby, if present), and the actual data points are 
    overlaid on top.
    
    Parameters:
    -----------
    obj: ns.Plate or pd.DataFrame object
        If DataFrame and `value = None`, the final column is used as the
        value.
    variable: str
        Column in DataFrame representing a timecourse-like variable,
        plotted on the x-axis.
    value_name: str
        Column in DataFrame representing the quantitative value,
        plotted on the y-axis.
    condition: str
        The names of one or more columns that specifies way the data
        is grouped for a single chart. Defaults to None.
    groupby:  str
        The names of one or more columns that further specify the
        way the data is grouped between different charts. Defaults
        to None.
    cmap: The colormap to use. Any Holoviews/Bokeh colormap is fine.
        Uses Holoviews default if None.
    show_all: bool
        If groupby is not None, whether or not to use a drop-down or
        to show all the plots (layout). Note that this can be pretty
        buggy from Holoview's layout system. There is usually a way
        to how all the info you want, in a nice way. Just play
        around.
    show_points: bool
        Shows all the data points. I don't even know why this is an
        argument. Default will show points if there are multiple
        replicates. Unless you have a really good reason, don't
        change this.
    legend: str
        First controls whether or not the legend is shown, then its
        position. Defaults to False, though 'top' would be a good
        option, or 'top_left' if using groupby.
    height: int
        The height of the chart.
    width: int
        The width of the chart.
    
    Returns:
    --------
    chart: the final Holoviews chart
    """
    # Get data and metadata
    df, locs, auto_value, case = _parse_data_obj(obj)
    if value_name is not None:
        check_df_col(df, value_name, 'value_name')
    else:
        value_name = auto_value
    
    # Check columns
    check_df_col(df, variable, 'variable')
    check_df_col(df, condition, 'condition')

    if not isinstance(groupby, (list, tuple)):
        groupby = [groupby]
    for group in groupby:
        check_df_col(df, group, 'groupby')

    groups = [elem for elem in (condition, *groupby) if elem is not None]
    replicates, df = aggregate_replicates(df, variable, value_name, groups)

    # Auto-colomapping
    if condition is None:
        n_conditions = 1
    else:
        n_conditions = len(df[condition].unique())
    if cmap == 'CategoryN':
        if n_conditions <= 10:
            cmap = 'Category10'
        else:
            cmap = 'Category20'

    # Determine if colormap is discrete or continuous
    cmap_len = max(bokeh.palettes.all_palettes[cmap].keys())
    discrete_cmap = True if cmap_len < 256 else False

    if isinstance(cmap, dict) and condition is not None:
        # Sorting in hv.Curve is, somehow always, alphabetical.
        # So, generate hv.Cycle object that way.
        line_color = hv.Cycle(
            [cmap[k] for k in sorted(df[condition].unique())]
        )
    else:
        # Use Palette to nicely sample from continuous colormap
        if not discrete_cmap:
            cmap = hv.Palette(cmap, samples=n_conditions+2).values[1:-1]
        line_color = hv.Cycle(cmap)

    # Pull out available encodings (column names)
    encodings = [*list(df.columns)]

    # Set options
    base_opts = dict(
        height=height,
        width=width,
        padding=0.1,
        show_legend=legend,
        color=condition,
    )

    line_opts = base_opts
    line_opts['color'] = line_color
    scat_opts = dict(size=6, fill_alpha=0.75, tools=['hover'], cmap=cmap)
    scat_opts.update(base_opts)

    # Now, start to actually make the chart
    points = hv.Scatter(
        df,
        variable,
        [value_name, *encodings]
    )

    lines = hv.Curve(
        df,
        variable,
        [(f'mean_{value_name}', value_name),
        *encodings]
    )

    if groups:
        points = points.groupby(groups)
        lines = lines.groupby(groups)

    # Output chart as desired
    if show_points is None:
        if replicates:
            chart = lines*points
        else:
            chart = lines
    elif show_points:
        chart = lines*points
    else:
        chart = lines

    # Overlay each line plot
    if condition is not None:
        chart = chart.overlay(condition)

    chart = chart.opts({
        'Curve': line_opts,
        'Scatter': scat_opts,
    }).opts(
        legend_position='right',
        toolbar='above',
    )

    # groupby among different charts
    if groupby is not None:
        # If groupby, show as side-by-side, or dropdown
        if show_all:
            chart = chart.layout(groupby)

    return chart



from functools import wraps

import pandas as pd
try:
    import pandas_flavor as pf
except ImportError:
    pf = None

import ninetysix as ns

# For placing viz plot methods on Plate
def _get_viz_attr(Plate, plot):
    attr = getattr(ns.viz, plot)
    @wraps(attr)
    def wrapper(*args, **kwargs):
        return attr(Plate, *args, **kwargs)

    return plot, wrapper

def _set_viz_attrs(Plate):
    plots = [func for func in dir(ns.viz) if 'plot_' in func]
    for plot in plots:
        attr_pair = _get_viz_attr(Plate, plot)
        setattr(Plate, *attr_pair)


# For placing pandas attributes/functions on Plate
# rather only being accessible via Plate.df.
# pandas_flavor (pf) helps return pd.DataFrame output as Plate
if pf is not None:
    @pf.register_dataframe_method
    def as_plate(df, locations, values):
        """Adds a method .as_plate() to wrap pd.DataFrame as ns.Plate"""
        # Create plate object with correct value_name
        plate = ns.Plate(df, value_name=values[-1])

        # Add locations
        plate.locations = locations

        # Add values
        plate.values = values
        
        return plate

    def _get_pandas_attrs(Plate, attr_name):
        """Creates wrappers for pandas functions to Plate.df"""
        attr = getattr(pd.DataFrame, attr_name)
        if callable(attr):
            @wraps(attr)
            def wrapper(*args, **kwargs):
                # head and tail are not used as new Plates; return pd obj is fine
                if attr_name in ('head', 'tail'):
                    output = attr(Plate.df, *args, **kwargs)

                # Default to index=False
                elif attr_name in ('to_csv', 'to_excel'):
                    if 'index' in kwargs:
                        index = kwargs.pop('index')
                    else:
                        index = False
                    output = attr(Plate.df, index=index, *args, **kwargs)

                # .as_plate() method returns DataFrame back as Plate object
                else:
                    output = attr(Plate.df, *args, **kwargs).as_plate(
                        Plate.locations, Plate.values
                    )
                
                return output
            
            attr_pair = (attr_name, wrapper)
        
        else:
            attr = getattr(Plate.df, attr_name)
            attr_pair = (attr_name, attr)

        return attr_pair

    def _set_pandas_attrs(Plate):
        """Assigns pandas attributes/methods to Plate from Plate.df"""
        _pd_attrs = dir(pd.DataFrame)
        _pd_deprecated = ['as_blocks', 'blocks', 'ftypes', 'is_copy', 'ix']
        _ns_incompat = ['values']
        ignore = [*_pd_deprecated, *_ns_incompat]
        
        for attr_name in _pd_attrs:
            if (attr_name in ignore) or (attr_name[0] == '_'):
                continue
            attr_pair = _get_pandas_attrs(Plate, attr_name)
            setattr(Plate, *attr_pair)



import pandas as pd

from ..parsers import well_regex, pad


def check_inputs(Plate):
    """Checks that all inputs are provided correctly.

    Returns True if all tests pass.
    """
    df = None
    if isinstance(Plate.data, pd.DataFrame):
        df = Plate.data

    if df is not None:
        message = 'column'

    elif isinstance(Plate.data, dict):
        df = pd.DataFrame(Plate.data)
        message = 'key'

    elif isinstance(Plate.data, list):
        df = pd.DataFrame(
            data=Plate.data,
            columns=['well', Plate.value_name]
        )

    well_cols = [col for col in df.columns if str(col).lower() == 'well']
    if 'well' not in [well.lower() for well in well_cols]:
        raise ValueError(
            f"No 'well' value found in your {message}s."
        )
    if len(well_cols) != 1:
        raise ValueError(
            "Multiple 'well' columns detected in your data."
        )

    well_col = well_cols[0]
    if (Plate.value_name is None) & (well_col == df.columns[-1]):
        raise ValueError(
            "Your final {message} is assumed to be your value {message},"
            "but found {message} related to well."
        )

    bad_types = [type(well) for well in df[well_col]
                 if not isinstance(well, str)]

    if bad_types:
        raise TypeError(
            f"Well values must be of type string, found: {set(bad_types)}"
        )

    if Plate.value_name is not None:
        if df.columns[-1] is not None and Plate.value_name not in df.columns:
            raise ValueError(
                f"'{Plate.value_name}' not present in your data, "\
                f"options are: {list(df.columns)}"
            )

    # TODO: Add row and column check?

    return True


def check_annotations(Plate, annotations):
    """Checks that all annotations are specified correctly,
    especially in the context of the Plate input.

    Returns annotation type if all tests pass.
    """
    if isinstance(annotations, dict):
        # Check for nested dict
        for column in annotations:
            if not isinstance(annotations[column], dict):
                raise ValueError(
                    "Annotations should be a nested dictionary, following the"\
                    " form {new_column_name: {wells: annotation}}. Your "\
                    "input did not contain a nested dictionary."
                )

        # Check keys
        acceptable_kwargs = ('default', 'standard', 'else', 'other')
        rows = list('ABCDEFGHIJKLMNOP')
        cols = [str(i) for i in range(1, 25)]
        acceptable_wells = [row+col for row in rows for col in cols]
        acceptable_wells += [pad(row+col) for row in rows for col in cols]
        acceptable_wells = set(acceptable_wells)
        
        for column in annotations:
            working_annotations = well_regex(annotations[column],
                                             padded=Plate.zero_padding)
            
            nonwell_keys = set(working_annotations.keys()) - acceptable_wells
            
            if len(nonwell_keys) == 0:
                continue

            if len(nonwell_keys) > 1:
                raise ValueError(
                    f"Multiple non-well keys found in '{column}' dict. "\
                    f'Provide only one of {acceptable_kwargs}.'
                )

            # Check identified key as default key
            default_key = list(nonwell_keys)[0]

            if default_key.lower() not in acceptable_kwargs:
                raise ValueError(
                    f"Non-well key ('{default_key}') found in '{column}' dict "\
                    f'that does not match acceptable default arguments '
                    f'({acceptable_kwargs}).'
                )
        # Assign type if passes
        annotation_type = dict
    
    elif isinstance(annotations, str):

        # TODO(check stuff)
        
        # Assign type if passes
        annotation_type = 'excel'

    else:
        raise NotImplementedError(
            'Only dictionary and excel-based mapping annotations are '
            'currently supported.'
        )

    return annotation_type


def check_df_col(df, column, name=None):
    """Checks for the presence of a column (or columns) in a tidy
    DataFrame with an informative error message. Passes silently,
    otherwise raises error.
    """
    if column is None:
        return
    
    if name is None:
        error_message = f"The value '{column}' is not present in any of your data's columns."
    else:
        error_message = f"Your {name} '{column}' is not present in any of your data's columns."
    error_message += "\nYou may be looking for:\n  " + \
        str(list(df.columns))

    if column not in df.columns:
        raise ValueError(error_message)



from setuptools import setup, find_packages


with open("README.md", "r") as f:
    long_description = f.read()


with open("ninetysix/__init__.py", "r") as f:
    init = f.readlines()

for line in init:
    if '__author__' in line:
        __author__ = line.split("'")[-2]
    if '__email__' in line:
        __email__ = line.split("'")[-2]
    if '__version__' in line:
        __version__ = line.split("'")[-2]

        
setup(
    name='ninetysix',
    version=__version__,
    author=__author__,
    author_email=__email__,
    description='A general package for tidying, annotating, and analyzing 96-well plate data.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://palmhjell.github.io/ninetysix/',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'pandas',
        'bokeh',
        'holoviews',
        'pandas_flavor',
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ),
)
